@@ -142,7 +142,7 @@ func (self *Handler) initialize() {
 func (self *Handler) generate_diff() {
 	self.diff_map = nil
 	jobs := make([]diff_job, 0, 32)
-	self.collection.Apply(func(path, typ, changed_path string) error {
+	_ = self.collection.Apply(func(path, typ, changed_path string) error {
 		if typ == "diff" {
 			if is_path_text(path) && is_path_text(changed_path) {
 				jobs = append(jobs, diff_job{path, changed_path})
@@ -184,11 +184,10 @@ func (self *Handler) highlight_all() {
 		self.async_results <- r
 		self.lp.WakeupMainThread()
 	}()
-
 }
 
 func (self *Handler) load_all_images() {
-	self.collection.Apply(func(path, item_type, changed_path string) error {
+	_ = self.collection.Apply(func(path, item_type, changed_path string) error {
 		if path != "" && is_image(path) {
 			image_collection.AddPaths(path)
 			self.image_count++
@@ -523,6 +522,27 @@ func (self *Handler) scroll_to_next_change(backwards bool) bool {
 	return false
 }
 
+func (self *Handler) scroll_to_next_file(backwards bool) bool {
+	if backwards {
+		for i := self.scroll_pos.logical_line - 1; i >= 0; i-- {
+			line := self.logical_lines.At(i)
+			if line.line_type == TITLE_LINE {
+				self.scroll_pos = ScrollPos{i, 0}
+				return true
+			}
+		}
+	} else {
+		for i := self.scroll_pos.logical_line + 1; i < self.logical_lines.Len(); i++ {
+			line := self.logical_lines.At(i)
+			if line.line_type == TITLE_LINE {
+				self.scroll_pos = ScrollPos{i, 0}
+				return true
+			}
+		}
+	}
+	return false
+}
+
 func (self *Handler) scroll_to_next_match(backwards, include_current_match bool) bool {
 	if self.current_search == nil {
 		return false
@@ -616,6 +636,8 @@ func (self *Handler) dispatch_action(name, args string) error {
 	case `scroll_to`:
 		done := false
 		switch {
+		case strings.Contains(args, "file"):
+			done = self.scroll_to_next_file(strings.Contains(args, `prev`))
 		case strings.Contains(args, `change`):
 			done = self.scroll_to_next_change(strings.Contains(args, `prev`))
 		case strings.Contains(args, `match`):
@@ -627,11 +649,9 @@ func (self *Handler) dispatch_action(name, args string) error {
 			}
 			done = self.scroll_lines(amt) != 0
 		default:
-			npos := self.scroll_pos
+			npos := ScrollPos{}
 			if strings.Contains(args, `end`) {
 				npos = self.max_scroll_pos
-			} else {
-				npos = ScrollPos{}
 			}
 			done = npos != self.scroll_pos
 			self.scroll_pos = npos
