@@ -14,7 +14,7 @@ typedef enum ScrollTypes { SCROLL_LINE = -999999, SCROLL_PAGE, SCROLL_FULL } Scr
 
 typedef struct {
     bool mLNM, mIRM, mDECTCEM, mDECSCNM, mDECOM, mDECAWM, mDECCOLM, mDECARM, mDECCKM,
-         mBRACKETED_PASTE, mFOCUS_TRACKING, mDECSACE, mHANDLE_TERMIOS_SIGNALS;
+         mBRACKETED_PASTE, mFOCUS_TRACKING, mDECSACE, mHANDLE_TERMIOS_SIGNALS, mINBAND_RESIZE_NOTIFICATION;
     MouseTrackingMode mouse_tracking_mode;
     MouseTrackingProtocol mouse_tracking_protocol;
 } ScreenModes;
@@ -97,6 +97,7 @@ typedef struct {
     struct {
         unsigned int cursor_x, cursor_y, scrolled_by;
         index_type lines, columns;
+        color_type cursor_bg;
     } last_rendered;
     bool is_dirty, scroll_changed, reload_all_gpu_data;
     Cursor *cursor;
@@ -132,6 +133,8 @@ typedef struct {
     union {
         struct {
             unsigned int redraws_prompts_at_all: 1;
+            unsigned int uses_special_keys_for_cursor_movement: 1;
+            unsigned int supports_click_events: 1;
         };
         unsigned int val;
     } prompt_settings;
@@ -226,6 +229,7 @@ void set_title(Screen *self, PyObject*);
 void desktop_notify(Screen *self, unsigned int, PyObject*);
 void set_icon(Screen *self, PyObject*);
 void set_dynamic_color(Screen *self, unsigned int code, PyObject*);
+void color_control(Screen *self, unsigned int code, PyObject*);
 void clipboard_control(Screen *self, int code, PyObject*);
 void shell_prompt_marking(Screen *self, char *buf);
 void file_transmission(Screen *self, PyObject*);
@@ -272,6 +276,7 @@ uint8_t screen_current_key_encoding_flags(Screen *self);
 void screen_report_key_encoding_flags(Screen *self);
 int screen_detect_url(Screen *screen, unsigned int x, unsigned int y);
 int screen_cursor_at_a_shell_prompt(const Screen *);
+bool screen_prompt_supports_click_events(const Screen *);
 bool screen_fake_move_cursor_to_position(Screen *, index_type x, index_type y);
 bool screen_send_signal_for_key(Screen *, char key);
 bool get_line_edge_colors(Screen *self, color_type *left, color_type *right);
